"""
Middleware to ensure best practices of DRF endpoints.
"""
import logging

from rest_framework_jwt.authentication import BaseJSONWebTokenAuthentication

from .permissions import JwtHasScope

log = logging.getLogger(__name__)


class EnsureJWTAuthSettingsMiddleware(object):
    """
    Django middleware object that ensures the proper Permission classes
    are set on all endpoints that use JWTAuthentication.
    """
    _required_permission_classes = (JwtHasScope,)
    _view_does_not_support_scopes = 'VIEW_DOES_NOT_SUPPORT_SCOPES'

    def _includes_base_class(self, iter_classes, base_class):
        """
        Returns whether any class in iter_class is a subclass of the given base_class.
        """
        return any(
            issubclass(auth_class, base_class) for auth_class in iter_classes,
        )

    def _add_missing_jwt_permission_classes(self, view_class):
        """
        Adds permissions classes that should exist for Jwt based authentication,
        if needed.
        """
        view_permission_classes = getattr(view_class, 'permission_classes', tuple())

        for perm_class in self._required_permission_classes:

            if not self._includes_base_class(view_permission_classes, perm_class):
                log.warning(
                    u"The view %s allows Jwt Authentication but needs to include the %s permission class.",
                    view_class.__name__,
                    perm_class.__name__,
                )

        view_class.permission_classes = view_permission_classes
        view_class.permission_classes += self._required_permission_classes

    def _add_missing_jwt_scopes(self, view_class):
        """
        Adds restricting scopes if none already exist for Jwt based authentication.
        """
        view_required_scopes = getattr(view_class, 'required_scopes', [])
        if not view_required_scopes:
            view_class.required_scopes = [self._view_does_not_support_scopes]

    def process_view(self, request, view_func, view_args, view_kwargs):  # pylint: disable=unused-argument
        view_class = getattr(view_func, 'view_class', view_func)

        view_authentication_classes = getattr(view_class, 'authentication_classes', tuple())
        if self._includes_base_class(view_authentication_classes, BaseJSONWebTokenAuthentication):
            self._add_missing_jwt_permission_classes(view_class)
            self._add_missing_jwt_scopes(view_class)



""" Permission classes. """
from django.core.exceptions import ImproperlyConfigured
from opaque_keys.edx.keys import CourseKey
from rest_framework.permissions import BasePermission
import waffle

from edx_rest_framework_extensions.authentication import is_jwt_authenticated
from edx_rest_framework_extensions.config import SWITCH_ENFORCE_JWT_SCOPES
from edx_rest_framework_extensions.jwt_decoder import decode_jwt_filters, decode_jwt_scopes


class IsSuperuser(BasePermission):
    """ Allows access only to superusers. """

    def has_permission(self, request, view):
        return request.user and request.user.is_superuser


class IsStaff(BasePermission):
    """
    Allows access to "global" staff users..
    """
    def has_permission(self, request, view):
        return request.user.is_staff


class IsUserInUrl(BasePermission):
    """
    Allows access if the requesting user matches the user in the URL.
    """
    def has_permission(self, request, view):
        user_parameter_name = 'username'
        url_username = (
            request.GET.get(user_parameter_name, '') or
            getattr(request, 'parser_context', {}).get('kwargs', {}).get(user_parameter_name, '')
        )
        return request.user.username.lower() == url_username.lower()


class IsJwtAuthenticated(BasePermission):
    """
    Returns whether the request was successfully authenticated with JwtAuthentication.
    """
    def has_permission(self, request, view):
        return is_jwt_authenticated(request)


class JwtHasScope(BasePermission):
    """
    The request is authenticated as a user and the token used has the right scope.
    """
    message = 'JWT missing required scopes.'

    def has_permission(self, request, view):
        if _should_skip_jwt_scope_enforcement(request):
            return True
        jwt_scopes = decode_jwt_scopes(request.auth)
        required_scopes = set(self.get_scopes(request, view))
        if required_scopes.issubset(jwt_scopes):
            return True
        return False

    def get_scopes(self, request, view):
        """
        Return the required scopes defined on the view.
        """
        try:
            return getattr(view, 'required_scopes')
        except AttributeError:
            raise ImproperlyConfigured('JwtHasScope requires the view to define the required_scopes attribute')


class JwtHasContentOrgFilterForRequestedCourse(BasePermission):
    """
    The JWT used to authenticate contains the appropriate content provider
    filter for the requested course resource.
    """
    message = 'JWT missing required content_org filter.'

    def has_permission(self, request, view):
        """
        Ensure that the course_id kwarg provided to the view contains one
        of the organizations specified in the content provider filters
        in the JWT used to authenticate.
        """
        if _should_skip_jwt_scope_enforcement(request):
            return True
        course_key = CourseKey.from_string(view.kwargs.get('course_id'))
        jwt_filters = decode_jwt_filters(request.auth)
        for provider_type, filter_value in jwt_filters:
            if provider_type == 'content_org' and filter_value == course_key.org:
                return True
        return False


def _should_skip_jwt_scope_enforcement(request):
    """
    Returns True if either the request is not JWT authenticated or if
    JWT scopes enforcement is disabled.
    """
    if not is_jwt_authenticated(request):
        return True
    return not waffle.switch_is_active(SWITCH_ENFORCE_JWT_SCOPES)
